# Copyright (c) 2018 Simon Guest
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import os
import os.path
import pwd
import re
import subprocess
import sys

class Mapper(object):
    def __init__(self):
        self._rpm_by_path = {}
        self._yum_repos_by_rpm = {}
        self._username = {}
        self._isfile = {}
        self._excluded = {}

    def username(self, uid):
        if uid not in self._username:
            try:
                pw = pwd.getpwuid(uid)
                name = pw[0]
            except KeyError:
                name = str(uid)
            self._username[uid] = name
        else:
            name = self._username[uid]
        return name

    def isfile(self, path):
        """Whether path is a regular file (or symlink to one) in the filesystem."""
        if path in self._isfile:
            isfile = self._isfile[path]
        else:
            isfile = os.path.isfile(path)
            self._isfile[path] = isfile
        return isfile

    def rpm(self, path):
        if path in self._rpm_by_path:
            package = self._rpm_by_path[path]
        else:
            package = None
            if os.path.exists(path):
                rpm = subprocess.Popen(["rpm", "-qf", "--qf", "%{NAME}\n", path], stdout = subprocess.PIPE, universal_newlines=True)
                line = rpm.stdout.readline().rstrip('\n')
                if not line.endswith('is not owned by any package'):
                    package = line
            self._rpm_by_path[path] = package
        return package

    def yum_repos(self, rpm):
        if rpm in self._yum_repos_by_rpm:
            repos = self._yum_repos_by_rpm[rpm]
        else:
            repos = set()
            yum = subprocess.Popen(["yum", "info", rpm], stdout = subprocess.PIPE, stderr = subprocess.PIPE, universal_newlines=True)
            for line in yum.stdout:
                m = re.match(r"""(From )?[Rr]epo\s*:\s*(\S*)""", line)
                if m:
                    repos.add(m.group(2))
            if len(repos) == 0:
                errorline = yum.stderr.read()
                if errorline.startswith('Error'):
                    sys.stderr.write('Mapper::yum_repo(%s) %s' % (rpm, errorline))
            self._yum_repos_by_rpm[rpm] = repos
        return repos

    def excluded(self, path, cls, config):
        """Look up config and return whether this path is excluded for cls."""
        if cls not in self._excluded:
            self._excluded[cls] = {}
        excluded_for_class = self._excluded[cls]
        if path in excluded_for_class:
            return excluded_for_class[path]
        else:
            if config.exclude_file(cls, path):
                excluded = True
            else:
                package = self.rpm(path)
                if package is None:
                    excluded = False
                elif config.include_rpm(cls, package):
                    excluded = False
                elif config.exclude_rpm(cls, package):
                    excluded = True
                else:
                    repos = self.yum_repos(package)
                    if len(repos) == 0:
                        excluded = False
                    elif config.exclude_any_yum_repos(cls, repos):
                        excluded = True
                    else:
                        excluded = False
            excluded_for_class[path] = excluded
        return excluded



# Copyright (c) 2018 Simon Guest
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

class KeyedReader(object):

    def __init__(self, path, keyfn):
        self._path = path
        self._f = open(path, 'r')
        self._keyfn = keyfn
        self.next()

    def __str__(self):
        return 'KeyedReader(%s)' % self._path

    def next(self):
        if self._f is not None:
            self.line = self._f.readline()
            if self.line != '':
                self.key = self._keyfn(self.line)
            else:
                self.key = None
                self._f.close()
                self._f = None



# Copyright (c) 2018 Simon Guest
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

class KeyedReaderTree(object):
    """A KeyedReaderTree is a binary tree used for merging lines from KeyReader's
    in order of key."""

    def __init__(self, reader=None):
        # Nodes with children have no reader themselves.
        # Also, the left child is filled first, so cannot be None if there is a child.
        self._reader = reader
        self._left = None
        self._right = None
        self._next = self._reader
        self.n = 1 if reader is not None else 0
        self.lastkey = None

    def __str__(self):
        return 'KRT(%d, %s, %s, %s)' % (self.n, self._reader, self._left, self._right)

    def _set_next(self):
        if self._reader is not None:
            if self._reader.key is not None:
                self._next = self._reader
            else:
                self._next = None
        elif self._right is None or self._right.key is None:
            if self._left.key is not None:
                self._next = self._left
            else:
                self._next = None
        elif self._left.key is None:
            self._next = self._right
        else:
            if self._left.key <= self._right.key:
                self._next = self._left
            else:
                self._next = self._right

    @property
    def key(self):
        if self._next is not None:
            return self._next.key
        else:
            return None

    @property
    def line(self):
        if self._next is not None:
            return self._next.line
        else:
            return ''

    def insert(self, reader):
        if self._reader is None and self._left is None and self._right is None:
            self._reader = reader
        elif self._reader is not None:
            self._left = KeyedReaderTree(self._reader)
            self._reader = None
            self.insert(reader)
        elif self._right is None:
            self._right = KeyedReaderTree(reader)
        else:
            if self._left.n < self._right.n:
                self._left.insert(reader)
            else:
                self._right.insert(reader)
        self.n += 1
        self._set_next()

    def lines(self):
        while self.key is not None:
            self.lastkey = self.key
            line = self.line
            yield line
            self.next()

    def next(self):
        if self._next is not None:
            self._next.next()
            self._set_next()

    def write(self, f, level=0):
        if self._reader is not None:
            f.write('%s%s\n' % ('.' * level, self._reader))
        if self._left is not None:
            f.write('%sL\n' % ('.' * level))
            self._left.write(f, level + 1)
        if self._right is not None:
            f.write('%sR\n' % ('.' * level))
            self._right.write(f, level + 1)



# Copyright (c) 2018 Simon Guest
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gzip
import os
import pendulum
import re
import sys

def get_tagged_fields(s):
    """Extract tagged snoopy fields as a dict."""
    # Alas embedded spaces cause difficulty, as snoopy doesn't quote them in the logfile.
    # We therefore split on a cunning regex
    cunningRE = re.compile(r"""(\s+[a-z]+:)""")
    cunningSplit = re.split(cunningRE, s)
    toks = cunningSplit[0].split(':', 1) + cunningSplit[1:]
    fields = {}
    for i in range(len(toks) // 2):
        tag = toks[2 * i].lstrip().rstrip(':')
        value = toks[2 * i + 1]
        fields[tag] = value
    if 'filename' not in fields:
        sys.stderr.write("warning: get_tagged_fields failed to find filename for %s\n" % s)
    return fields

class Reader(object):
    def __init__(self, log, logfile_dt, config):
        self._config = config
        self._logfile_dt = logfile_dt
        self._logpath = os.path.join(self._config.logdir, log)

    def collate_to(self, collator):
        loglineRE = re.compile(r"""^(\S+\s+\d+\s+\d+:\d+:\d+)\s+(\S+)\s+\S+\[(\d+)\]:\s+\[([^\]]*)\]:\s+(.*)$""")
        logf = gzip.open(self._logpath)
        loglineno = 0
        try:
            for logline_bytes in logf:
                try:
                    logline = logline_bytes.decode('utf-8')
                    loglineno += 1
                    m = loglineRE.match(logline)
                    if m:
                        # infer the year for the timestamp, which is usually the same as the logfile year,
                        # except when we roll over from Dec to Jan
                        timestamp_s = m.group(1)
                        timestamp_year = self._logfile_dt.year - 1 if timestamp_s.startswith('Dec') and self._logfile_dt.month == 1 else self._logfile_dt.year
                        timestamp = pendulum.parse('%d %s' % (timestamp_year, timestamp_s), tz=pendulum.now().timezone, strict=False)
                        fields = get_tagged_fields(m.group(4))
                        command = m.group(5).rstrip()
                        collator.command(timestamp, fields, command)
                    else:
                        sys.stderr.write('warning: ignoring badly formatted line at %s:%d\n' % (self._logpath, loglineno))
                except UnicodeDecodeError:
                    sys.stderr.write('warning: ignoring badly encoded line at %s:%d\n' % (self._logpath, loglineno))
        except:
            sys.stderr.write('failed at %s:%d\n' % (self._logpath, loglineno))
            raise
        finally:
            logf.close()



# Copyright (c) 2018 Simon Guest
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import os.path
import pendulum
import re
import sys

from .Collator import Collator
from .Config import Config
from .Mapper import Mapper
from .Reader import Reader

class Scanner(object):

    def __init__(self, args):
        self._args = args
        self._config = Config(args)
        self._mapper = Mapper()
        self._collator = Collator(self._config, self._mapper)

    def _get_last_collation(self):
        timestampRE = re.compile(r"""^(\d\d\d\d)(\d\d)(\d\d)$""")
        try:
            with open(self._config.last_collation_file) as f:
                m = timestampRE.match(f.read())
                if m:
                    return pendulum.DateTime(int(m.group(1)), int(m.group(2)), int(m.group(3)), tzinfo=pendulum.now().timezone)
            return None
        except (IOError, ValueError):
            return None

    def _set_last_collation(self, dt):
        with open(self._config.last_collation_file, 'w') as f:
            f.write('%s\n' % dt.strftime('%Y%m%d'))

    def scan(self):
        last_collation_dt = self._get_last_collation()

        # important to process logfiles in order, so timestamps are preserved
        for entry in sorted(os.listdir(self._config.logdir)):
            snoopyLogRE = re.compile(r"""^snoopy-(\d\d\d\d)(\d\d)(\d\d).gz$""")
            m = snoopyLogRE.match(entry)
            if m:
                logfile_year = int(m.group(1))
                logfile_month = int(m.group(2))
                logfile_day = int(m.group(3))
                logfile_dt = pendulum.DateTime(logfile_year, logfile_month, logfile_day, tzinfo=pendulum.now().timezone)
                if last_collation_dt is None or last_collation_dt < logfile_dt:
                    reader = Reader(entry, logfile_dt, self._config)
                    if self._args.verbose:
                        sys.stdout.write('collating %s\n' % entry)
                    reader.collate_to(self._collator)
                    last_collation_dt = logfile_dt
                    self._set_last_collation(last_collation_dt)
                else:
                    if self._args.verbose:
                        sys.stdout.write('skipping %s\n' % entry)



# Copyright (c) 2018 Simon Guest
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import os.path
import sys

from .Config import Config
from .Mapper import Mapper
from .KeyedReader import KeyedReader
from .KeyedReaderTree import KeyedReaderTree
from .util import bare_hostname, append_and_set_timestamp, timestamp_from_str

class PostProcessor(object):

    def __init__(self, args):
        self._config = Config(args)
        self._mapper = Mapper()

    def _get_collated_files(self, cls, host, paths):
        collationdir = self._config.host_collation_dir(cls, host)
        n = len(collationdir)
        for root, dirs, files in os.walk(collationdir):
            for filename in files:
                path = os.path.join(root, filename)[n:]
                if path != '/.collated':
                    if path not in paths:
                        paths[path] = set()
                    paths[path].add(host)

    def list_packages(self, classes):
        paths = {}
        for cls in classes if len(classes) > 0 else ['all']:
            self._get_collated_files(cls, bare_hostname(), paths)
        for path in sorted(paths.keys()):
            package = self._mapper.rpm(path)
            repos = self._mapper.yum_repos(package) if package is not None else None
            if package is not None:
                print('%s:%s:%s' % (path, str(package), str(repos)))

    def list_files(self, classes):
        paths = {}
        for cls in classes if len(classes) > 0 else ['all']:
            for host in self._config.collated_hosts(cls):
                self._get_collated_files(cls, host, paths)
        for path in sorted(paths.keys()):
            print('%s %s' % (path, ','.join(sorted(list(paths[path])))))

    def list_excluded(self, classes, purge=False):
        paths = {}
        for cls in classes if len(classes) > 0 else ['all']:
            root = self._config.localhost_collation_dir(cls)
            self._get_collated_files(cls, bare_hostname(), paths)
            for path in paths:
                if self._mapper.excluded(path, cls, self._config):
                    if purge:
                        filepath = os.path.join(root, os.path.relpath(path, '/'))
                        print('rm %s' % filepath)
                        os.remove(filepath)
                    else:
                        print(path)
            if purge:
                self._purge_empty_dirs(cls)

    def _purge_empty_dirs(self, cls):
        for root, dirs, files in os.walk(self._config.localhost_collation_dir(cls), topdown=False):
            for dirname in dirs:
                dirpath = os.path.join(root, dirname)
                try:
                    os.rmdir(dirpath)
                    print('rmdir %s ' % dirpath)
                except:
                    pass

    @classmethod
    def timestamp(cls, line):
        """Return just the timestamp from a line in a collated file."""
        return timestamp_from_str(line.split(maxsplit=1)[0])

    def consolidate(self):
        for cls in self._config.classes:
            hosts = self._config.collated_hosts(cls)
            all_relpaths = {}
            for host in hosts:
                collationdir = self._config.host_collation_dir(cls, host)
                n = len(collationdir) + 1
                for root, dirs, files in os.walk(collationdir):
                    for filename in files:
                        relpath = os.path.join(root, filename)[n:]
                        if relpath not in all_relpaths:
                            all_relpaths[relpath] = []
                        all_relpaths[relpath].append(host)
            for relpath, hosts in all_relpaths.items():
                krt = KeyedReaderTree()
                inpaths = [os.path.join(self._config.host_collation_dir(cls, host), relpath) for host in hosts]
                outpath = os.path.join(self._config.consolidation_dir(cls), relpath)
                for inpath in inpaths:
                    krt.insert(KeyedReader(inpath, self.__class__.timestamp))
                if os.path.exists(outpath):
                    krt.insert(KeyedReader(outpath, self.__class__.timestamp))
                os.makedirs(os.path.dirname(outpath), exist_ok=True)
                outpathnew = '%s.new' % outpath
                with open(outpathnew, 'w') as f:
                    for line in krt.lines():
                        f.write(line)
                os.rename(outpathnew, outpath)
                # set the timestamp according to the last key
                t = krt.lastkey.int_timestamp
                os.utime(outpath, (t, t))
            self._finalize_consolidated(cls)

    def _finalize_consolidated(self, cls):
        """Ensure the collated files don't get consolidated again, by moving them."""
        hosts = self._config.collated_hosts(cls)
        all_relpaths = {}
        for host in hosts:
            collationdir = self._config.host_collation_dir(cls, host)
            n = len(collationdir) + 1
            for root, dirs, files in os.walk(collationdir, topdown=False):
                for filename in files:
                    relpath = os.path.join(root, filename)[n:]
                    inpath = os.path.join(collationdir, relpath)
                    outpath = os.path.join(self._config.consolidation_dir(cls, host), relpath)
                    os.makedirs(os.path.dirname(outpath), exist_ok=True)
                    if not os.path.exists(outpath):
                        os.rename(inpath, outpath)
                    else:
                        append_and_set_timestamp(inpath, outpath)
                        os.remove(inpath)
                # remove all the directories, which should be empty now
                # if not, it's because someone else is busy writing here,
                # so ignore that for now, and we'll pick it up next time
                try:
                    os.rmdir(root)
                except OSError:
                    pass



#!/usr/bin/env python
#
# Copyright (c) 2018 Simon Guest
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import sys

from snoopy_log_collator.Config import ConfigError
from snoopy_log_collator.PostProcessor import PostProcessor
from snoopy_log_collator.Scanner import Scanner
from snoopy_log_collator.version import get_version

def main():
    parser = argparse.ArgumentParser(description='collate snoopy logfiles')
    parser.add_argument('-v', '--verbose', action='store_true', help='verbose output')
    parser.add_argument('-c', '--config', metavar='FILE', help='configuration file')
    parser.add_argument('command', choices=['collate','consolidate','list-files','list-packages','list-excluded','purge-excluded','version'], help='command to run')
    parser.add_argument('args', nargs=argparse.REMAINDER, help='command arguments')
    args = parser.parse_args()

    try:
        if args.command == 'version':
            print('snoopy-log-collator v%s' % get_version())
        elif args.command == 'list-files':
            p = PostProcessor(args)
            p.list_files(args.args)
        elif args.command == 'list-packages':
            p = PostProcessor(args)
            p.list_packages(args.args)
        elif args.command == 'list-excluded':
            p = PostProcessor(args)
            p.list_excluded(args.args)
        elif args.command == 'purge-excluded':
            p = PostProcessor(args)
            p.list_excluded(args.args, purge=True)
        elif args.command == 'consolidate':
            p = PostProcessor(args)
            p.consolidate()
        elif args.command == 'collate':
            scanner = Scanner(args)
            scanner.scan()
    except ConfigError as e:
        sys.stderr.write('%s\n' % e)
        sys.exit(1)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# distutils setup script for filebutler package

from setuptools import setup, find_packages

long_description = """Snoopy-log-collator collates the log files from snoopy, producing a
summary of what programs have been run on the system, when, and by whom.  The
output comprises a number of files, one per command invoked, each being a list
of invocation instances.
"""

setup(name='snoopy-log-collator',
      use_scm_version=True,
      setup_requires=['setuptools_scm', 'setuptools_scm_git_archive'],
      description='Collate logfiles from snoopy logger',
      long_description=long_description,
      author='Simon Guest',
      author_email='simon.guest@tesujimath.org',
      url='https://github.com/tesujimath/snoopy-log-collator',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 3',
          'Topic :: System :: Systems Administration',
          'Topic :: Utilities',
      ],
      packages=find_packages(),
      entry_points={
        'console_scripts': [
            'snoopy-log-collator = snoopy_log_collator.__main__:main',
        ],
      },
      package_data={
          'doc': ['*.rst'],
      },
      license='GPLv3',
      install_requires=[
          'pendulum>=2',
          'pytoml',
          'setuptools',
      ],
      python_requires='>=3',
     )



# Copyright (c) 2018 Simon Guest
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path
import pendulum
import re

from .util import bare_hostname, timestamp_str

class Collator(object):
    def __init__(self, config, mapper):
        self._config = config
        self._mapper = mapper
        self._hostname = bare_hostname()

    def _outpath(self, cls, filename):
        if os.path.isabs(filename):
            path = os.path.join(self._config.localhost_collation_dir(cls), filename[1:])
        else:
            path = os.path.join(self._config.localhost_collation_dir(cls), filename)
        return os.path.normpath(path)

    def command(self, timestamp, fields, command):
        filename = fields['filename']
        if os.path.isabs(filename):
            filepath = filename
        else:
            filepath = os.path.normpath(os.path.join(fields['cwd'], filename))
        user = self._mapper.username(int(fields['uid']))
        if self._mapper.isfile(filepath):
            for cls in self._config.classes_with_all:
                if not self._mapper.excluded(filepath, cls, self._config):
                    outpath = self._outpath(cls, filepath)
                    if not os.path.exists(outpath):
                        os.makedirs(os.path.dirname(outpath), exist_ok=True)
                    with open(outpath, 'a') as outf:
                        outf.write('%s %s %s %s\n' % (timestamp_str(timestamp), self._hostname, user, command))
                    t = timestamp.int_timestamp
                    os.utime(outpath, (t, t))



# Copyright (c) 2018 Simon Guest
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pendulum

def bare_hostname():
    """Hostname without domain."""
    return os.uname()[1].split('.')[0]

def append_and_set_timestamp(inpath, outpath):
    with open(inpath, 'r') as inf:
        with open(outpath, 'a') as outf:
            done = False
            bufsize = 104857600 # 10MB at a time
            bytes = inf.read(bufsize)
            while bytes != '':
                outf.write(bytes)
                bytes = inf.read(bufsize)
    t = os.stat(inpath).st_mtime
    os.utime(outpath, (t, t))

def timestamp_str(t0):
    return t0.strftime('%Y%m%d-%H:%M:%S')

def timestamp_from_str(s):
    return pendulum.from_format(s, 'YYYYMMDD-HH:mm:ss', tz=pendulum.now().timezone)



# Copyright (c) 2018 Simon Guest
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path
import pytoml as toml
import re
import sys

from .util import bare_hostname

def expand(s):
    return os.path.expanduser(os.path.expandvars(s))

class ConfigError(Exception):

    def __init__(self, filename, msg):
        self.filename = filename
        self.msg = msg

    def __str__(self):
        return('Configuration error %s: %s' % (self.filename, self.msg))

class Config(object):

    def __init__(self, args):
        self._filename = None
        if args.config:
            self._filename = args.config
            if not os.path.exists(self._filename):
                raise ConfigError('%s' % self._filename, 'file not found')
        else:
            attempt_files = [
                os.path.expanduser('~/.snoopy-log-collator.toml'),
                '/etc/snoopy-log-collator.toml',
            ]
            for attempt in attempt_files:
                if os.path.exists(attempt):
                    self._filename = attempt
                    break
            if self._filename is None:
                raise ConfigError('', 'none of %s found' % ', '.join(attempt_files))

        with open(self._filename, 'rb') as f:
            try:
                self._config = toml.load(f)
            except toml.TomlError as e:
                raise ConfigError(self._filename, 'TOML error at line %d, %s' % (e.line, e.message))
        self._validate()

    def _validate(self):
        if 'class' in self._config and 'all' in self._config['class']:
            raise ConfigError(self._filename, 'invalid class "all"')

    def localhost_collation_dir(self, cls):
        return os.path.join(expand(self._config['collation-dir']), cls, bare_hostname())

    def host_collation_dir(self, cls, hostname):
        return os.path.join(expand(self._config['collation-dir']), cls, hostname)

    def consolidation_dir(self, cls, host=None):
        subdir = 'ALL' if host is None else host
        return os.path.join(expand(self._config['consolidation-dir']), cls, subdir)

    @property
    def last_collation_file(self):
        return os.path.join(expand(self._config['collation-dir']), '.%s.collated' % bare_hostname())

    @property
    def logdir(self):
        return expand(self._config['log-dir'])

    @property
    def classes(self):
        return list(self._config['class'])

    @property
    def classes_with_all(self):
        # add the pseudo 'all' class
        return ['all'] + list(self._config['class'])

    def collated_hosts(self, cls):
        return os.listdir(os.path.join(expand(self._config['collation-dir']), cls))

    def _config_for_class(self, cls):
        if 'class' not in self._config or cls not in self._config['class']:
            return None
        else:
            return self._config['class'][cls]

    def exclude_file(self, cls, name):
        config_for_class = self._config_for_class(cls)
        if config_for_class is not None and 'exclude' in config_for_class:
            exclude = config_for_class['exclude']
            if 'file' in exclude:
                for name_re in exclude['file']:
                    if re.search(name_re, name):
                        return True
                return False
            else:
                return False
        else:
            return False

    def include_rpm(self, cls, name):
        config_for_class = self._config_for_class(cls)
        if config_for_class is not None and 'include' in config_for_class:
            include = config_for_class['include']
            if 'rpm' in include:
                for name_re in include['rpm']:
                    if re.search(name_re, name):
                        return True
                return False
            else:
                return False
        else:
            return False

    def exclude_rpm(self, cls, name):
        config_for_class = self._config_for_class(cls)
        if config_for_class is not None and 'exclude' in config_for_class:
            exclude = config_for_class['exclude']
            if 'rpm' in exclude:
                for name_re in exclude['rpm']:
                    if re.search(name_re, name):
                        return True
                return False
            else:
                return False
        else:
            return False

    def exclude_any_yum_repos(self, cls, names):
        config_for_class = self._config_for_class(cls)
        if config_for_class is not None and 'exclude' in config_for_class:
            exclude = config_for_class['exclude']
            if 'yum-repo' in exclude:
                for name_re in exclude['yum-repo']:
                    for name in names:
                        if re.search(name_re, name):
                            return True
                return False
            else:
                return False
        else:
            return False

